/*
The software is provided by the Institute of Commercial Cryptography Standards 
(ICCS), and is used for algorithm submissions in the Next-generation Commercial 
Cryptographic Algorithms Program. 

ICCS doesn't represent or warrant that the operation of the software will be 
uninterrupted or error-free in all cases. ICCS will take no responsibility for 
the use of the software or the results thereof, if the software is used for any 
other purposes.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "drng.h"
#include "SIG_AlgorithmInstance.h"

#define SEED_LENGTH_BYTE 64

#define KAT_SIG_SUCCESS 0
#define KAT_SIG_FOPEN_ERROR -1
#define KAT_SIG_CRYPTO_FAILURE -3

void fprintstr(FILE *file_output, unsigned char *msg, unsigned long long len);

// DRNG_ctx for generating pseudorandom numbers within the SIG scheme
DRNG_ctx drng_algorithm;

/****************output KAT_SIG_AlgorithmInstance.txt****************/
int main()
{
	unsigned char *nonce1;
	unsigned char *nonce2;
	// DRNG_ctx for generating seed
	DRNG_ctx drng_seed;
	// DRNG_ctx for generating message
	DRNG_ctx drng_msg;
	char algoname_output[128];
	FILE *file_output;
	unsigned char *seed, *m, *sn, *pk, *sk;
	unsigned long long pk_byts, sk_byts, sn_byts;
	int m_byts = 56;
	int rtn;
	// init drng_seed using nonce1
	nonce1 = (unsigned char *)malloc(SEED_LENGTH_BYTE);
	for (int i = 0; i < SEED_LENGTH_BYTE / 4; i++)
	{
		memcpy(nonce1 + 4 * i, "seed", 4);
	}
	init_random_number(&drng_seed, nonce1, SEED_LENGTH_BYTE);
	// init drng_msg using nonce2
	nonce2 = (unsigned char *)malloc(SEED_LENGTH_BYTE);
	for (int i = 0; i < SEED_LENGTH_BYTE / 3; i++)
	{
		memcpy(nonce2 + 3 * i, "msg", 3);
	}
	memcpy(nonce2 + SEED_LENGTH_BYTE - 1, "m", 1);
	init_random_number(&drng_msg, nonce2, SEED_LENGTH_BYTE);
	// open KAT_SIG_AlgorithmInstance.txt
	sprintf(algoname_output, "KAT_SIG_%s.txt", ALGORITHM_INSTANCE);
	if ((file_output = fopen(algoname_output, "wb")) == NULL)
		return KAT_SIG_FOPEN_ERROR;

	pk_byts = sig_get_pk_byts();
	sk_byts = sig_get_sk_byts();
	sn_byts = sig_get_sn_byts();
	pk = (unsigned char *)calloc(pk_byts, sizeof(unsigned char));
	sk = (unsigned char *)calloc(sk_byts, sizeof(unsigned char));
	sn = (unsigned char *)calloc(sn_byts, sizeof(unsigned char));
	seed = (unsigned char *)calloc(SEED_LENGTH_BYTE, sizeof(unsigned char));
	m = (unsigned char *)calloc(128, sizeof(unsigned char));
	for (int i = 0; i < 10; i++)
	{
		fprintf(file_output, "count = %d\n", i);
		// generate seed using drng_seed
		get_random_number(&drng_seed, seed, SEED_LENGTH_BYTE * 8);
		// generate message using drng_msg
		get_random_number(&drng_msg, m, m_byts * 8);
		// print seed
		fprintf(file_output, "seedlen = %08X\n", SEED_LENGTH_BYTE);
		fprintf(file_output, "seed = ");
		fprintstr(file_output, seed, SEED_LENGTH_BYTE);
		fprintf(file_output, "\n");

		/****************the SIG scheme starts here****************/
		// init drng_algorithm using seed
		init_random_number(&drng_algorithm, seed, SEED_LENGTH_BYTE);

		/****************       Key generate       ****************/
		rtn = sig_keygen(pk, &pk_byts, sk, &sk_byts);
		if (rtn)
		{
			printf("sig_keygen returned %d\n", rtn);
			return KAT_SIG_CRYPTO_FAILURE;
		}
		// print public key
		if (OUTPUT_BLANK_TEST_VECTORS)
			fprintf(file_output, "pklen = \n");
		else
			fprintf(file_output, "pklen = %08llX\n", pk_byts);
		fprintf(file_output, "pk = ");
		fprintstr(file_output, pk, pk_byts);
		fprintf(file_output, "\n");
		// print private key
		if (OUTPUT_BLANK_TEST_VECTORS)
			fprintf(file_output, "sklen = \n");
		else
			fprintf(file_output, "sklen = %08llX\n", sk_byts);
		fprintf(file_output, "sk = ");
		fprintstr(file_output, sk, sk_byts);
		fprintf(file_output, "\n");
		// print message
		fprintf(file_output, "mlen = %08X\n", m_byts);
		fprintf(file_output, "m = ");
		fprintstr(file_output, m, m_byts);
		fprintf(file_output, "\n");

		/****************            Sign          ****************/
		rtn = sig_sign(sk, sk_byts, m, m_byts, sn, &sn_byts);
		if (rtn)
		{
			printf("sig_sign returned %d\n", rtn);
			return KAT_SIG_CRYPTO_FAILURE;
		}
		// print signature
		if (OUTPUT_BLANK_TEST_VECTORS)
			fprintf(file_output, "snlen = \n");
		else
			fprintf(file_output, "snlen = %08llX\n", sn_byts);
		fprintf(file_output, "sn = ");
		fprintstr(file_output, sn, sn_byts);
		fprintf(file_output, "\n");

		/****************           Verify         ****************/
		rtn = sig_verify(pk, pk_byts, sn, sn_byts, m, m_byts);
		if (rtn)
		{
			printf("sig_verify returned %d\n", rtn);
			return KAT_SIG_CRYPTO_FAILURE;
		}
		fprintf(file_output, "\n");
		m_byts += 8;
	}
	free(m);
	free(seed);
	free(sn);
	free(sk);
	free(pk);
	free(nonce2);
	free(nonce1);
	// close KAT_SIG_AlgorithmInstance.txt
	fclose(file_output);
	return KAT_SIG_SUCCESS;
}

void fprintstr(FILE *file_output, unsigned char *msg, unsigned long long len)
{
	for (unsigned long long i = 0; i < len; i++)
		fprintf(file_output, "%02X", msg[i]);
}