/*
The software is provided by the Institute of Commercial Cryptography Standards 
(ICCS), and is used for algorithm submissions in the Next-generation Commercial 
Cryptographic Algorithms Program. 

ICCS doesn't represent or warrant that the operation of the software will be 
uninterrupted or error-free in all cases. ICCS will take no responsibility for 
the use of the software or the results thereof, if the software is used for any 
other purposes.
*/

#ifndef KAT_KEX_H
#define KAT_KEX_H

// Set "OUTPUT_BLANK_TEST_VECTORS" as 0 to generate test vector files rather than blank template
#define OUTPUT_BLANK_TEST_VECTORS 1

// Set "ALGORITHM_INSTANCE" as your algorithm instance name (no more than 64 bytes)
#define ALGORITHM_INSTANCE "AlgorithmInstance"

#ifdef __cplusplus
extern "C"
{
#endif

	/// @brief Obtain the claimed round
	/// @return Claimed round
	unsigned long long kex_get_rounds();

	/// @brief Obtain the claimed byte length of the long-term public key
	/// @return If the authentication function is available, return the claimed byte length of the long-term public key; if not, return 0
	unsigned long long kex_get_pk_byts();

	/// @brief Obtain the claimed byte length of the long-term private key
	/// @return If the authentication function is available, return the claimed byte length of the long-term private key; if not, return 0
	unsigned long long kex_get_sk_byts();

	/// @brief Obtain the maximum byte length of the initiator cached data
	/// @return The maximum byte length of the initiator cached data
	unsigned long long kex_get_pa_byts();

	/// @brief Obtain the maximum byte length of the responder cached data
	/// @return The maximum byte length of the responder cached data
	unsigned long long kex_get_pb_byts();

	/// @brief Obtain the claimed byte length of the shared secret key
	/// @return Claimed byte length of the shared secret key
	unsigned long long kex_get_ss_byts();

	/// @brief Obtain the claimed byte length of the exchanging message
	/// @return Claimed byte length of the exchanging message
	unsigned long long kex_get_bw_byts();

	/// @brief initialize by the initiator
	/// @param pka Initiator long-term public key
	/// @param pka_byts Byte length of the initiator long-term public key (0 if no long-term public key)
	/// @param ska Initiator long-term private key
	/// @param ska_byts Byte length of the initiator long-term private key (0 if no long-term private key)
	/// @param pa Initiator cached data
	/// @param pa_byts Byte length of the initiator cached data
	/// @return If run successfully, return 0; otherwise, return -1 or a self-defined negative error code
	int kex_init_a(
		unsigned char *pka, unsigned long long *pka_byts,
		unsigned char *ska, unsigned long long *ska_byts,
		unsigned char *pa, unsigned long long *pa_byts);

	/// @brief initialize by the responder
	/// @param pkb Responder long-term public key
	/// @param pkb_byts Byte length of the responder long-term public key (0 if no long-term public key)
	/// @param skb Responder long-term private key
	/// @param skb_byts Byte length of the responder long-term private key (0 if no long-term private key)
	/// @param pb Responder cached data
	/// @param pb_byts Byte length of the responder cached data
	/// @return If run successfully, return 0; otherwise, return -1 or a self-defined negative error code
	int kex_init_b(
		unsigned char *pkb, unsigned long long *pkb_byts,
		unsigned char *skb, unsigned long long *skb_byts,
		unsigned char *pb, unsigned long long *pb_byts);

	/// @brief send by the initiator
	/// @param pkb Responder long-term public key
	/// @param pkb_byts Byte length of the responder long-term public key
	/// @param pa Initiator cached data (might be updated after execution)
	/// @param pa_byts Byte length of the initiator cached data
	/// @param ma Message sent by the initiator
	/// @param ma_byts Byte length of the message sent by the initiator
	/// @return If run successfully, return 0; otherwise, return -1 or a self-defined negative error code
	int kex_send_a(
		unsigned char *pkb, unsigned long long pkb_byts,
		unsigned char *pa, unsigned long long *pa_byts,
		unsigned char *ma, unsigned long long *ma_byts);

	/// @brief send by the responder (if round = 2, generate the shared secret key in this function)
	/// @param pka Initiator long-term public key
	/// @param pka_byts Byte length of the initiator long-term public key (0 if no long-term public key)
	/// @param ma Message sent by the initiator
	/// @param ma_byts Byte length of the message sent by the initiator
	/// @param pb Responder cached data (might be updated after execution)
	/// @param pb_byts Byte length of the responder cached data
	/// @param mb Message sent by the the responder
	/// @param mb_byts Byte length of the message sent by the responder
	/// @param ssb Shared secret key of the responder
	/// @param ssb_byts Byte length of the shared secret key of the responder
	/// @return If run successfully, return 0; otherwise, return -1 or a self-defined negative error code
	int kex_send_b(
		unsigned char *pka, unsigned long long pka_byts,
		unsigned char *ma, unsigned long long ma_byts,
		unsigned char *pb, unsigned long long *pb_byts,
		unsigned char *mb, unsigned long long *mb_byts,
		unsigned char *ssb, unsigned long long *ssb_byts);

	/// @brief respond by the initiator (generate the shared secret key)
	/// @param mb Message sent by the responder
	/// @param mb_byts Byte length of the message sent by the responder
	/// @param pa Initiator cached data (might be updated after execution)
	/// @param pa_byts Byte length of the initiator cached data
	/// @param ma Message sent by the initiator (if round = 3, generate a response message)
	/// @param ma_byts Byte length of the message sent by the initiator
	/// @param ssa Shared secret key of the initiator
	/// @param ssa_byts Byte length of the shared secret key of the initiator
	/// @return If run successfully, return 0; otherwise, return -1 or a self-defined negative error code
	int kex_resp_a(
		unsigned char *mb, unsigned long long mb_byts,
		unsigned char *pa, unsigned long long *pa_byts,
		unsigned char *ma, unsigned long long *ma_byts,
		unsigned char *ssa, unsigned long long *ssa_byts);

	/// @brief respond by the responder (if round = 3, generate the shared secret key in this function)
	/// @param ma Message sent by the initiator
	/// @param ma_byts Byte length of the message sent by the initiator
	/// @param pb Responder cached data (might be updated after execution)
	/// @param pb_byts Byte length of the responder cached data
	/// @param mb Message sent by the responder (if round > 3, generate a response message)
	/// @param mb_byts Byte length of the message sent by the responder
	/// @param ssb Shared secret key of the responder
	/// @param ssb_byts Byte length of the shared secret key of the responder
	/// @return If run successfully, return 0; otherwise, return -1 or a self-defined negative error code
	int kex_resp_b(
		unsigned char *ma, unsigned long long ma_byts,
		unsigned char *pb, unsigned long long *pb_byts,
		unsigned char *mb, unsigned long long *mb_byts,
		unsigned char *ssb, unsigned long long *ssb_byts);

#ifdef __cplusplus
}
#endif
#endif