/*
The software is provided by the Institute of Commercial Cryptography Standards 
(ICCS), and is used for algorithm submissions in the Next-generation Commercial 
Cryptographic Algorithms Program. 

ICCS doesn't represent or warrant that the operation of the software will be 
uninterrupted or error-free in all cases. ICCS will take no responsibility for 
the use of the software or the results thereof, if the software is used for any 
other purposes.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "auxfunc.h"

#define HASH_SUCCESS 0
#define MEMORY_ALLOCATION_FAILED -2
#define HASH_FAILED -3
#define INVALID_DIGESTBITLEN -4

#define L1 512    //GB/T 15852.2-2012 (HMAC)
#define L2 256    //GB/T 15852.2-2012 (HMAC)

#define FF1(x, y, z) ((x) ^ (y) ^ (z))
#define FF2(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define GG1(x, y, z) ((x) ^ (y) ^ (z))
#define GG2(x, y, z) ((((y) ^ (z)) & (x)) ^ (z))
#define L_SHIFT(a, n) ((a) << (n) | ((a) & 0xFFFFFFFF) >> (32 - (n)))
#define P0(x) ((x) ^ L_SHIFT((x), 9) ^ L_SHIFT((x), 17))
#define P1(x) ((x) ^ L_SHIFT((x), 15) ^ L_SHIFT((x), 23))
#define PUT32(a, b) ((a)[0] = (unsigned char)((b) >> 24), \
					 (a)[1] = (unsigned char)((b) >> 16), \
					 (a)[2] = (unsigned char)((b) >> 8),  \
					 (a)[3] = (unsigned char)(b))

static void sm3_bit_init(unsigned int *init_digest)
{
	init_digest[0] = 0x7380166F;
	init_digest[1] = 0x4914B2B9;
	init_digest[2] = 0x172442D7;
	init_digest[3] = 0xDA8A0600;
	init_digest[4] = 0xA96F30BC;
	init_digest[5] = 0x163138AA;
	init_digest[6] = 0xE38DEE4D;
	init_digest[7] = 0xB0FB0E4E;
}

static void sm3_bit_compress(unsigned int dgst[8], const unsigned char *msg, unsigned long long blocks)
{
	unsigned int A, B, C, D, E, F, G, H;
	unsigned int W[68];
	unsigned int W_prime[64];
	unsigned int SS1, SS2, TT1, TT2;
	int i;
	while (blocks--)
	{
		/***********************Round function*************************/
		for (i = 0; i < 16; i++)
		{
			W[i] = ((unsigned int)(msg + i * 4)[0] << 24 | (unsigned int)(msg + i * 4)[1] << 16 | (unsigned int)(msg + i * 4)[2] << 8 | (unsigned int)(msg + i * 4)[3]);
		}
		for (; i < 68; i++)
		{
			W[i] = P1(W[i - 16] ^ W[i - 9] ^ L_SHIFT(W[i - 3], 15)) ^ L_SHIFT(W[i - 13], 7) ^ W[i - 6];
		}
		for (i = 0; i < 64; i++)
		{
			W_prime[i] = W[i] ^ W[i + 4];
		}
		A = dgst[0];
		B = dgst[1];
		C = dgst[2];
		D = dgst[3];
		E = dgst[4];
		F = dgst[5];
		G = dgst[6];
		H = dgst[7];
		for (i = 0; i < 64; i++)
		{
			if (i < 16)
			{
				SS1 = L_SHIFT(L_SHIFT(A, 12) + E + L_SHIFT(0x79cc4519U, i & 0x1F), 7);
			}
			else
			{
				SS1 = L_SHIFT(L_SHIFT(A, 12) + E + L_SHIFT(0x7a879d8aU, i & 0x1F), 7);
			}
			SS2 = SS1 ^ L_SHIFT(A, 12);
			if (i < 16)
			{
				TT1 = FF1(A, B, C) + D + SS2 + W_prime[i];
				TT2 = GG1(E, F, G) + H + SS1 + W[i];
			}
			else
			{
				TT1 = FF2(A, B, C) + D + SS2 + W_prime[i];
				TT2 = GG2(E, F, G) + H + SS1 + W[i];
			}
			D = C;
			C = L_SHIFT(B, 9);
			B = A;
			A = TT1;
			H = G;
			G = L_SHIFT(F, 19);
			F = E;
			E = P0(TT2);
		}
		dgst[0] ^= A;
		dgst[1] ^= B;
		dgst[2] ^= C;
		dgst[3] ^= D;
		dgst[4] ^= E;
		dgst[5] ^= F;
		dgst[6] ^= G;
		dgst[7] ^= H;
		msg += 64;
	}
}

static void sm3_bit(const unsigned char *msg, unsigned long long msg_bitlen, unsigned char *dgst)
{
	unsigned long long block_num;
	unsigned long long remain;
	block_num = msg_bitlen / 512;
	remain = msg_bitlen & 0x1FF;
	/**********************Initializing value**********************/
	unsigned int digest[8];
	sm3_bit_init(digest);
	/************************Round function************************/
	if (block_num != 0)
	{
		sm3_bit_compress(digest, msg, block_num);
	}
	unsigned char block[64];
	memset(block, 0, 64);
	memcpy(block, msg + block_num * 64, (remain + 7) >> 3);
	block[remain >> 3] &= ((0xFF00 >> (remain & 0x7)) & 0xFF);
	block[remain >> 3] |= (1 << (7 - (remain & 0x7)));
	/***************************Padding****************************/
	if (remain <= 512 - 65)
	{
		memset(block + (remain >> 3) + 1, 0, (512 - remain - 65) >> 3);
	}
	else
	{
		memset(block + (remain >> 3) + 1, 0, (512 - remain - 1) >> 3);
		sm3_bit_compress(digest, block, 1);
		memset(block, 0, 64 - 8);
	}
	PUT32(block + 56, block_num >> 32 << 9);
	PUT32(block + 60, (block_num << 9) + (remain));
	/************************Round function************************/
	sm3_bit_compress(digest, block, 1);
	/**********************Output hash value***********************/
	for (int i = 0; i < 8; i++)
	{
		PUT32(dgst + i * 4, digest[i]);
	}
}

// key = SM3(UTF-8("3.14159265358979")) || SM3(UTF-8("2.71828182845904"))
static const unsigned char key[] = {
	0x53, 0x07, 0xf6, 0xd5, 0xeb, 0x6a, 0x3c, 0xed,
	0x3d, 0x24, 0xc5, 0x3c, 0xc9, 0xc8, 0x2c, 0xce,
	0x2f, 0x89, 0x36, 0x39, 0x70, 0x23, 0xf0, 0x69,
	0x5c, 0x26, 0xc8, 0x0c, 0x1a, 0xb1, 0x82, 0xa7,
	0x1d, 0xb0, 0x2b, 0xa9, 0x2f, 0x54, 0x40, 0x18,
	0x11, 0x5a, 0x96, 0xe7, 0x19, 0x66, 0x2c, 0xa3,
	0x2b, 0x7c, 0x7e, 0xfc, 0x0a, 0x6d, 0x24, 0x82,
	0x15, 0x07, 0x66, 0xba, 0x6f, 0x65, 0x5b, 0x8e};

static unsigned long long key_bitlen = (sizeof(key) / sizeof(key[0])) * 8;

static void normalize(unsigned char *input, unsigned long long total_bits)
{
	unsigned long long full_bytes = total_bits / 8;
	unsigned long long remaining_bits = total_bits % 8;
	if (remaining_bits > 0)
	{
		input[full_bytes] = input[full_bytes] & ~((1 << (8 - remaining_bits)) - 1);
	}
}

static int hmac(const unsigned char *message, unsigned long long message_bitlen, const unsigned char *key, unsigned long long key_bitlen, unsigned char *mac)
{
	// key expansion
	unsigned char K[L1 / 8] = {0};
	memcpy(K, key, key_bitlen / 8);
	unsigned char K1[L1 / 8], K2[L1 / 8];
	unsigned char IPAD[L1 / 8], OPAD[L1 / 8];
	memset(IPAD, 0x36, sizeof(IPAD));
	memset(OPAD, 0x5C, sizeof(OPAD));
	for (unsigned long long i = 0; i < L1 / 8; i++)
	{
		K1[i] = K[i] ^ IPAD[i];
		K2[i] = K[i] ^ OPAD[i];
	}
	// hash operation
	unsigned char H1[L2 / 8], H2[L2 / 8];
	unsigned char *temp1;
	temp1 = (unsigned char *)malloc(L1 / 8 + (message_bitlen + 7) / 8);
	if (temp1 == NULL)
	{
		fprintf(stderr, "ERROR: Memory allocation failed at %s, line %d. \n", __FILE__, __LINE__);
		return MEMORY_ALLOCATION_FAILED;
	}
	memcpy(temp1, K1, L1 / 8);
	memcpy(temp1 + L1 / 8, message, (message_bitlen + 7) / 8);
	sm3_bit(temp1, L1 + message_bitlen, H1);
	// output transformation
	unsigned char temp2[L1 / 8 + L2 / 8];
	memcpy(temp2, K2, L1 / 8);
	memcpy(temp2 + L1 / 8, H1, L2 / 8);
	sm3_bit(temp2, L1 + L2, H2);
	// truncation operation
	memcpy(mac, H2, L2 / 8);
	free(temp1);
	return HASH_SUCCESS;
}

static int pseudohash_512(const unsigned char *message, unsigned long long message_bitlen, unsigned char *hash512)
{
	// k1
	unsigned char k1[32];
	unsigned char *cascade_0200_message;
	cascade_0200_message = (unsigned char *)malloc((message_bitlen + 7) / 8 + 2);
	if (cascade_0200_message == NULL)
	{
		fprintf(stderr, "ERROR: Memory allocation failed at %s, line %d. \n", __FILE__, __LINE__);
		return MEMORY_ALLOCATION_FAILED;
	}
	cascade_0200_message[0] = 0x02;
	cascade_0200_message[1] = 0x00;
	memcpy(cascade_0200_message + 2, message, (message_bitlen + 7) / 8);
	if (hmac(cascade_0200_message, 16 + message_bitlen, key, key_bitlen, k1) != 0)
	{
		fprintf(stderr, "ERROR: HMAC failed at %s, line %d. \n", __FILE__, __LINE__);
		return HASH_FAILED;
	}
	// h1
	unsigned char h1[32];
	unsigned char *cascade_message_0200;
	cascade_message_0200 = (unsigned char *)malloc((message_bitlen + 7) / 8 + 2);
	if (cascade_message_0200 == NULL)
	{
		fprintf(stderr, "ERROR: Memory allocation failed at %s, line %d. \n", __FILE__, __LINE__);
		return MEMORY_ALLOCATION_FAILED;
	}
	memcpy(cascade_message_0200, message, (message_bitlen + 7) / 8);
	if (message_bitlen % 8 == 0)
	{
		cascade_message_0200[(message_bitlen + 7) / 8] = 0x02;
		cascade_message_0200[(message_bitlen + 7) / 8 + 1] = 0x00;
	}
	else
	{
		normalize(cascade_message_0200, message_bitlen);
		cascade_message_0200[(message_bitlen + 7) / 8 - 1] = cascade_message_0200[(message_bitlen + 7) / 8 - 1] ^ (0x02 >> (message_bitlen % 8));
		cascade_message_0200[(message_bitlen + 7) / 8] = (0x02 << (8 - message_bitlen % 8)) & 0xff;
		cascade_message_0200[(message_bitlen + 7) / 8 + 1] = 0x00;
	}
	sm3_bit(cascade_message_0200, 16 + message_bitlen, h1);
	// h2
	unsigned char h2[32];
	unsigned char k1_cascade_h1[64];
	memcpy(k1_cascade_h1, k1, 32);
	memcpy(k1_cascade_h1 + 32, h1, 32);
	sm3_bit(k1_cascade_h1, (sizeof(k1_cascade_h1) / sizeof(k1_cascade_h1)[0]) * 8, h2);
	// final result
	memcpy(hash512, h1, 32);
	memcpy(hash512 + 32, h2, 32);
	// free
	free(cascade_0200_message);
	free(cascade_message_0200);
	return HASH_SUCCESS;
}

static int pseudohash_768(const unsigned char *message, unsigned long long message_bitlen, unsigned char *hash768)
{
	// k1
	unsigned char k1[32];
	unsigned char *cascade_0300_message;
	cascade_0300_message = (unsigned char *)malloc((message_bitlen + 7) / 8 + 2);
	if (cascade_0300_message == NULL)
	{
		fprintf(stderr, "ERROR: Memory allocation failed at %s, line %d. \n", __FILE__, __LINE__);
		return MEMORY_ALLOCATION_FAILED;
	}
	cascade_0300_message[0] = 0x03;
	cascade_0300_message[1] = 0x00;
	memcpy(cascade_0300_message + 2, message, (message_bitlen + 7) / 8);
	if (hmac(cascade_0300_message, 16 + message_bitlen, key, key_bitlen, k1) != 0)
	{
		fprintf(stderr, "ERROR: HMAC failed at %s, line %d. \n", __FILE__, __LINE__);
		return HASH_FAILED;
	}
	// h1
	unsigned char h1[32];
	unsigned char *cascade_message_0300;
	cascade_message_0300 = (unsigned char *)malloc((message_bitlen + 7) / 8 + 2);
	if (cascade_message_0300 == NULL)
	{
		fprintf(stderr, "ERROR: Memory allocation failed at %s, line %d. \n", __FILE__, __LINE__);
		return MEMORY_ALLOCATION_FAILED;
	}
	memcpy(cascade_message_0300, message, (message_bitlen + 7) / 8);

	if (message_bitlen % 8 == 0)
	{
		cascade_message_0300[(message_bitlen + 7) / 8] = 0x03;
		cascade_message_0300[(message_bitlen + 7) / 8 + 1] = 0x00;
	}
	else
	{
		normalize(cascade_message_0300, message_bitlen);
		cascade_message_0300[(message_bitlen + 7) / 8 - 1] = cascade_message_0300[(message_bitlen + 7) / 8 - 1] ^ (0x03 >> (message_bitlen % 8));
		cascade_message_0300[(message_bitlen + 7) / 8] = (0x03 << (8 - message_bitlen % 8)) & 0xff;
		cascade_message_0300[(message_bitlen + 7) / 8 + 1] = 0x00;
	}
	sm3_bit(cascade_message_0300, 16 + message_bitlen, h1);
	// k2
	unsigned char k2[32];
	if (hmac(h1, (sizeof(h1) / sizeof(h1[0])) * 8, key, key_bitlen, k2) != 0)
	{
		fprintf(stderr, "ERROR: HMAC failed at %s, line %d. \n", __FILE__, __LINE__);
		return HASH_FAILED;
	}
	// h2
	unsigned char h2[32];
	unsigned char k1_cascade_h1[64];
	memcpy(k1_cascade_h1, k1, 32);
	memcpy(k1_cascade_h1 + 32, h1, 32);
	sm3_bit(k1_cascade_h1, (sizeof(k1_cascade_h1) / sizeof(k1_cascade_h1[0])) * 8, h2);
	// h3
	unsigned char h3[32];
	unsigned char h2_cascade_k2[64];
	memcpy(h2_cascade_k2, h2, 32);
	memcpy(h2_cascade_k2 + 32, k2, 32);
	sm3_bit(h2_cascade_k2, (sizeof(h2_cascade_k2) / sizeof(h2_cascade_k2[0])) * 8, h3);
	// final result
	memcpy(hash768, h1, 32);
	memcpy(hash768 + 32, h2, 32);
	memcpy(hash768 + 64, h3, 32);
	// free
	free(cascade_0300_message);
	free(cascade_message_0300);
	return HASH_SUCCESS;
}

static int pseudohash_1024(const unsigned char *message, unsigned long long message_bitlen, unsigned char *hash1024)
{
	// k1
	unsigned char k1[32];
	unsigned char *cascade_0400_message;
	cascade_0400_message = (unsigned char *)malloc((message_bitlen + 7) / 8 + 2);
	if (cascade_0400_message == NULL)
	{
		fprintf(stderr, "ERROR: Memory allocation failed at %s, line %d. \n", __FILE__, __LINE__);
		return MEMORY_ALLOCATION_FAILED;
	}
	cascade_0400_message[0] = 0x04;
	cascade_0400_message[1] = 0x00;
	memcpy(cascade_0400_message + 2, message, (message_bitlen + 7) / 8);
	if (hmac(cascade_0400_message, 16 + message_bitlen, key, key_bitlen, k1) != 0)
	{
		fprintf(stderr, "ERROR: HMAC failed at %s, line %d. \n", __FILE__, __LINE__);
		return HASH_FAILED;
	}
	// h1
	unsigned char h1[32];
	unsigned char *cascade_message_0400;
	cascade_message_0400 = (unsigned char *)malloc((message_bitlen + 7) / 8 + 2);
	if (cascade_message_0400 == NULL)
	{
		fprintf(stderr, "ERROR: Memory allocation failed at %s, line %d. \n", __FILE__, __LINE__);
		return MEMORY_ALLOCATION_FAILED;
	}
	memcpy(cascade_message_0400, message, (message_bitlen + 7) / 8);
	normalize(cascade_message_0400, message_bitlen);
	if (message_bitlen % 8 == 0)
	{
		cascade_message_0400[(message_bitlen + 7) / 8] = 0x04;
		cascade_message_0400[(message_bitlen + 7) / 8 + 1] = 0x00;
	}
	else
	{
		cascade_message_0400[(message_bitlen + 7) / 8 - 1] = cascade_message_0400[(message_bitlen + 7) / 8 - 1] ^ (0x04 >> (message_bitlen % 8));
		cascade_message_0400[(message_bitlen + 7) / 8] = (0x04 << (8 - message_bitlen % 8)) & 0xff;
		cascade_message_0400[(message_bitlen + 7) / 8 + 1] = 0x00;
	}
	sm3_bit(cascade_message_0400, 16 + message_bitlen, h1);
	// k2
	unsigned char k2[32];
	if (hmac(h1, (sizeof(h1) / sizeof(h1[0])) * 8, key, key_bitlen, k2) != 0)
	{
		fprintf(stderr, "ERROR: HMAC failed at %s, line %d. \n", __FILE__, __LINE__);
		return HASH_FAILED;
	}
	// h2
	unsigned char h2[32];
	unsigned char k1_cascade_h1[64];
	memcpy(k1_cascade_h1, k1, 32);
	memcpy(k1_cascade_h1 + 32, h1, 32);
	sm3_bit(k1_cascade_h1, (sizeof(k1_cascade_h1) / sizeof(k1_cascade_h1[0])) * 8, h2);
	// k3
	unsigned char k3[32];
	if (hmac(h2, (sizeof(h2) / sizeof(h2[0])) * 8, key, key_bitlen, k3) != 0)
	{
		fprintf(stderr, "ERROR: HMAC failed at %s, line %d. \n", __FILE__, __LINE__);
		return HASH_FAILED;
	}
	// h3
	unsigned char h3[32];
	unsigned char h2_cascade_k2[64];
	memcpy(h2_cascade_k2, h2, 32);
	memcpy(h2_cascade_k2 + 32, k2, 32);
	sm3_bit(h2_cascade_k2, (sizeof(h2_cascade_k2) / sizeof(h2_cascade_k2[0])) * 8, h3);
	// k4
	unsigned char k4[32];
	if (hmac(h3, (sizeof(h3) / sizeof(h3[0])) * 8, key, key_bitlen, k4) != 0)
	{
		fprintf(stderr, "ERROR: HMAC failed at %s, line %d. \n", __FILE__, __LINE__);
		return HASH_FAILED;
	}
	// h4
	unsigned char h4[32];
	unsigned char k3_cascade_h3_k4[96];
	memcpy(k3_cascade_h3_k4, k3, 32);
	memcpy(k3_cascade_h3_k4 + 32, h3, 32);
	memcpy(k3_cascade_h3_k4 + 64, k4, 32);
	sm3_bit(k3_cascade_h3_k4, (sizeof(k3_cascade_h3_k4) / sizeof(k3_cascade_h3_k4[0])) * 8, h4);
	// final result
	memcpy(hash1024, h1, 32);
	memcpy(hash1024 + 32, h2, 32);
	memcpy(hash1024 + 64, h3, 32);
	memcpy(hash1024 + 96, h4, 32);
	// free
	free(cascade_0400_message);
	free(cascade_message_0400);
	return HASH_SUCCESS;
}

int sm3hash(int digestbitlen, const unsigned char *message, unsigned long long messagebitlen, unsigned char *digest)
{
	if (digestbitlen == 256)
	{
		sm3_bit(message, messagebitlen, digest);
		return HASH_SUCCESS;
	}
	else
	{
		fprintf(stderr, "ERROR: Digestbitlen not in the limitation. \n");
		return INVALID_DIGESTBITLEN;
	}
}

int pseudohash(int digestbitlen, const unsigned char *message, unsigned long long messagebitlen, unsigned char *digest)
{
	if (digestbitlen == 512)
	{
		return pseudohash_512(message, messagebitlen, digest);
	}
	else if (digestbitlen == 768)
	{
		return pseudohash_768(message, messagebitlen, digest);
	}
	else if (digestbitlen == 1024)
	{
		return pseudohash_1024(message, messagebitlen, digest);
	}
	else
	{
		fprintf(stderr, "ERROR: Digestbitlen not in the limitation. \n");
		return INVALID_DIGESTBITLEN;
	}
}

int pseudoXOF(int digestbitlen, const unsigned char *message, unsigned long long messagebitlen, unsigned char *digest)
{
	if (0 < digestbitlen && digestbitlen <= 256)
	{
		sm3_bit(message, messagebitlen, digest);
		normalize(digest, digestbitlen);
		return HASH_SUCCESS;
	}
	else if (256 < digestbitlen && digestbitlen <= 512)
	{
		if (pseudohash_512(message, messagebitlen, digest) == 0)
		{
			normalize(digest, digestbitlen);
			return HASH_SUCCESS;
		}
		else
		{
			fprintf(stderr, "ERROR: Execution error in %s. \n", __func__);
			return HASH_FAILED;
		}
	}
	else if (512 < digestbitlen && digestbitlen <= 768)
	{
		if (pseudohash_768(message, messagebitlen, digest) == 0)
		{
			normalize(digest, digestbitlen);
			return HASH_SUCCESS;
		}
		else
		{
			fprintf(stderr, "ERROR: Execution error in %s. \n", __func__);
			return HASH_FAILED;
		}
	}
	else if (768 < digestbitlen && digestbitlen <= 1024)
	{
		if (pseudohash_1024(message, messagebitlen, digest) == 0)
		{
			normalize(digest, digestbitlen);
			return HASH_SUCCESS;
		}
		else
		{
			fprintf(stderr, "ERROR: Execution error in %s. \n", __func__);
			return HASH_FAILED;
		}
	}
	else
	{
		fprintf(stderr, "ERROR: Digestbitlen not in the limitation. \n");
		return INVALID_DIGESTBITLEN;
	}
}